"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const SHOULD_RECONNECT_FALSE_MESSAGE = 'Provided shouldReconnect() returned false. Closing permanently.';
const SHOULD_RECONNECT_PROMISE_FALSE_MESSAGE = 'Provided shouldReconnect() resolved to false. Closing permanently.';
class AblativeWebSocket {
    constructor(url, protocolsOrOptions, options = {}) {
        this.url = url;
        this.onclose = null;
        this.onerror = null;
        this.onmessage = null;
        this.onopen = null;
        this.ondown = null;
        this.onreopen = null;
        this.CONNECTING = AblativeWebSocket.CONNECTING;
        this.OPEN = AblativeWebSocket.OPEN;
        this.CLOSING = AblativeWebSocket.CLOSING;
        this.CLOSED = AblativeWebSocket.CLOSED;
        this.hasBeenOpened = false;
        this.isClosed = false;
        this.messageBuffer = [];
        this.nextRetryTime = 0;
        this.reconnectCount = 0;
        this.lastKnownExtensions = '';
        this.lastKnownProtocol = '';
        this.listeners = {};
        if (protocolsOrOptions == null ||
            typeof protocolsOrOptions === 'string' ||
            Array.isArray(protocolsOrOptions)) {
            this.protocols = protocolsOrOptions;
        }
        else {
            options = protocolsOrOptions;
        }
        this.options = applyDefaultOptions(options);
        if (!this.options.wsConstructor) {
            if (typeof WebSocket !== 'undefined') {
                this.options.wsConstructor = WebSocket;
            }
            else {
                throw new Error('WebSocket not present in global scope and no ' +
                    'wsConstructor option was provided.');
            }
        }
        this.openNewWebSocket();
    }
    get binaryType() {
        return this.binaryTypeInternal || 'blob';
    }
    set binaryType(binaryType) {
        this.binaryTypeInternal = binaryType;
        if (this.ws) {
            this.ws.binaryType = binaryType;
        }
    }
    get bufferedAmount() {
        let sum = this.ws ? this.ws.bufferedAmount : 0;
        let hasUnknownAmount = false;
        this.messageBuffer.forEach((data) => {
            const byteLength = getDataByteLength(data);
            if (byteLength != null) {
                sum += byteLength;
            }
            else {
                hasUnknownAmount = true;
            }
        });
        if (hasUnknownAmount) {
            this.debugLog('Some buffered data had unknown length. bufferedAmount()' +
                ' return value may be below the correct amount.');
        }
        return sum;
    }
    get extensions() {
        return this.ws ? this.ws.extensions : this.lastKnownExtensions;
    }
    get protocol() {
        return this.ws ? this.ws.protocol : this.lastKnownProtocol;
    }
    get readyState() {
        return this.isClosed ? AblativeWebSocket.CLOSED : AblativeWebSocket.OPEN;
    }
    close(code, reason) {
        this.disposeSocket(code, reason);
        this.shutdown();
        this.debugLog('WebSocket permanently closed by client.');
    }
    send(data) {
        if (this.ws && this.ws.readyState === this.OPEN) {
            this.ws.send(data);
        }
        else {
            this.messageBuffer.push(data);
        }
    }
    reconnect() {
        if (this.isClosed) {
            throw new Error('Cannot call reconnect() on socket which is permanently closed.');
        }
        this.disposeSocket(1000, 'Client requested reconnect.');
        this.handleClose(undefined);
    }
    addEventListener(type, listener) {
        if (!this.listeners[type]) {
            this.listeners[type] = [];
        }
        this.listeners[type].push(listener);
    }
    dispatchEvent(event) {
        return this.dispatchEventOfType(event.type, event);
    }
    removeEventListener(type, listener) {
        if (this.listeners[type]) {
            this.listeners[type] = this.listeners[type].filter((l) => l !== listener);
        }
    }
    openNewWebSocket() {
        if (this.isClosed) {
            return;
        }
        const { connectTimeout, wsConstructor } = this.options;
        this.debugLog(`Opening new WebSocket to ${this.url}.`);
        const ws = new wsConstructor(this.url, this.protocols);
        ws.onclose = (event) => this.handleClose(event);
        ws.onerror = (event) => this.handleError(event);
        ws.onmessage = (event) => this.handleMessage(event);
        ws.onopen = (event) => this.handleOpen(event);
        this.connectTimeoutId = setTimeout(() => {
            // If this is running, we still haven't opened the websocket.
            // Kill it so we can try again.
            this.clearConnectTimeout();
            this.disposeSocket();
            this.handleClose(undefined);
        }, connectTimeout);
        this.ws = ws;
    }
    handleOpen(event) {
        if (!this.ws || this.isClosed) {
            return;
        }
        const { allClearResetTime } = this.options;
        this.debugLog('WebSocket opened.');
        if (this.binaryTypeInternal != null) {
            this.ws.binaryType = this.binaryTypeInternal;
        }
        else {
            this.binaryTypeInternal = this.ws.binaryType;
        }
        this.clearConnectTimeout();
        if (this.hasBeenOpened) {
            this.dispatchEventOfType('reopen', event);
        }
        else {
            this.dispatchEventOfType('open', event);
            this.hasBeenOpened = true;
        }
        this.messageBuffer.forEach((message) => this.send(message));
        this.messageBuffer = [];
        this.allClearTimeoutId = setTimeout(() => {
            this.clearAllClearTimeout();
            this.nextRetryTime = 0;
            this.reconnectCount = 0;
            const openTime = (allClearResetTime / 1000) | 0;
            this.debugLog(`WebSocket remained open for ${openTime} seconds. Resetting` +
                ' retry time and count.');
        }, allClearResetTime);
    }
    handleMessage(event) {
        if (this.isClosed) {
            return;
        }
        this.dispatchEventOfType('message', event);
    }
    handleClose(event) {
        if (this.isClosed) {
            return;
        }
        const { maxReconnectAttempts, shouldReconnect } = this.options;
        this.clearConnectTimeout();
        this.clearAllClearTimeout();
        if (this.ws) {
            this.lastKnownExtensions = this.ws.extensions;
            this.lastKnownProtocol = this.ws.protocol;
            this.ws = undefined;
        }
        this.dispatchEventOfType('down', event);
        if (this.reconnectCount >= maxReconnectAttempts) {
            this.stopReconnecting(event, this.getTooManyFailedReconnectsMessage());
            return;
        }
        const willReconnect = !event || shouldReconnect(event);
        if (typeof willReconnect === 'boolean') {
            this.handleWillReconnect(willReconnect, event, SHOULD_RECONNECT_FALSE_MESSAGE);
        }
        else {
            willReconnect.then((willReconnectResolved) => {
                if (this.isClosed) {
                    return;
                }
                this.handleWillReconnect(willReconnectResolved, event, SHOULD_RECONNECT_PROMISE_FALSE_MESSAGE);
            });
        }
    }
    handleError(event) {
        this.dispatchEventOfType('error', event);
        this.debugLog('WebSocket encountered an error.');
    }
    handleWillReconnect(willReconnect, event, denialReason) {
        if (willReconnect) {
            this.reestablishConnection();
        }
        else {
            this.stopReconnecting(event, denialReason);
        }
    }
    reestablishConnection() {
        const { minReconnectDelay, maxReconnectDelay, reconnectBackoffFactor } = this.options;
        this.reconnectCount++;
        const retryTime = this.nextRetryTime;
        this.nextRetryTime = Math.max(minReconnectDelay, Math.min(this.nextRetryTime * reconnectBackoffFactor, maxReconnectDelay));
        setTimeout(() => this.openNewWebSocket(), retryTime);
        const retryTimeSeconds = (retryTime / 1000) | 0;
        this.debugLog(`WebSocket was closed. Re-opening in ${retryTimeSeconds} seconds.`);
    }
    stopReconnecting(event, debugReason) {
        this.debugLog(debugReason);
        this.shutdown();
        if (event) {
            this.dispatchEventOfType('close', event);
        }
    }
    shutdown() {
        this.isClosed = true;
        this.clearAllTimeouts();
        this.messageBuffer = [];
    }
    disposeSocket(closeCode, reason) {
        if (!this.ws) {
            return;
        }
        // Use noop handlers instead of null because some WebSocket
        // implementations, such as the one from isomorphic-ws, raise a stink on
        // unhandled events.
        this.ws.onerror = noop;
        this.ws.onclose = noop;
        this.ws.onmessage = noop;
        this.ws.onopen = noop;
        this.ws.close(closeCode, reason);
        this.ws = undefined;
    }
    clearAllTimeouts() {
        this.clearConnectTimeout();
        this.clearAllClearTimeout();
    }
    clearConnectTimeout() {
        if (this.connectTimeoutId != null) {
            clearTimeout(this.connectTimeoutId);
            this.connectTimeoutId = undefined;
        }
    }
    clearAllClearTimeout() {
        if (this.allClearTimeoutId != null) {
            clearTimeout(this.allClearTimeoutId);
            this.allClearTimeoutId = undefined;
        }
    }
    dispatchEventOfType(type, event) {
        switch (type) {
            case 'close':
                if (this.onclose) {
                    this.onclose(event);
                }
                break;
            case 'error':
                if (this.onerror) {
                    this.onerror(event);
                }
                break;
            case 'message':
                if (this.onmessage) {
                    this.onmessage(event);
                }
                break;
            case 'open':
                if (this.onopen) {
                    this.onopen(event);
                }
                break;
            case 'down':
                if (this.ondown) {
                    this.ondown(event);
                }
                break;
            case 'reopen':
                if (this.onreopen) {
                    this.onreopen(event);
                }
                break;
        }
        if (type in this.listeners) {
            this.listeners[type]
                .slice()
                .forEach((listener) => this.callListener(listener, event));
        }
        return !event || !event.defaultPrevented;
    }
    callListener(listener, event) {
        if (typeof listener === 'function') {
            listener.call(this, event);
        }
        else {
            listener.handleEvent.call(this, event);
        }
    }
    debugLog(message) {
        if (this.options.debug) {
            // tslint:disable-next-line:no-console
            console.log(message);
        }
    }
    getTooManyFailedReconnectsMessage() {
        const { maxReconnectAttempts } = this.options;
        return `Failed to reconnect after ${maxReconnectAttempts} ${pluralize('attempt', maxReconnectAttempts)}. Closing permanently.`;
    }
}
exports.default = AblativeWebSocket;
AblativeWebSocket.DEFAULT_OPTIONS = {
    allClearResetTime: 5000,
    connectTimeout: 5000,
    debug: false,
    minReconnectDelay: 1000,
    maxReconnectDelay: 30000,
    maxReconnectAttempts: Number.POSITIVE_INFINITY,
    reconnectBackoffFactor: 1.5,
    shouldReconnect: () => true,
    wsConstructor: undefined,
};
AblativeWebSocket.CONNECTING = 0;
AblativeWebSocket.OPEN = 1;
AblativeWebSocket.CLOSING = 2;
AblativeWebSocket.CLOSED = 3;
function applyDefaultOptions(options) {
    const result = {};
    Object.keys(AblativeWebSocket.DEFAULT_OPTIONS).forEach((key) => {
        const value = options[key];
        result[key] =
            value === undefined
                ? AblativeWebSocket.DEFAULT_OPTIONS[key]
                : value;
    });
    return result;
}
function getDataByteLength(data) {
    if (typeof data === 'string') {
        // UTF-16 strings use two bytes per character.
        return 2 * data.length;
    }
    else if (data instanceof ArrayBuffer) {
        return data.byteLength;
    }
    else if (data instanceof Blob) {
        return data.size;
    }
    else {
        return undefined;
    }
}
function pluralize(s, n) {
    return n === 1 ? s : `${s}s`;
}
function noop() {
    // Nothing.
}
//# sourceMappingURL=ablative.js.map